"""
Magic Sets [1] rewriting implementation for Datalog.

[1] F. Bancilhon, D. Maier, Y. Sagiv, J. D. Ullman, in ACM PODS ’86, pp. 1–15.
"""

from abc import ABC, abstractmethod
from typing import Iterable, List, Set, Tuple


from ..config import config
from ..expressions import Constant, Expression, Symbol
from ..expression_walker import ExpressionWalker
from ..expression_pattern_matching import add_match
from ..logic import Negation
from ..probabilistic.expressions import ProbabilisticQuery
from . import extract_logic_predicates, DatalogProgram
from .exceptions import (
    BoundAggregationApplicationError,
    NegationInMagicSetsRewriteError,
    NonConjunctiveAntecedentInMagicSetsError,
    NoConstantPredicateFoundError,
)
from .expressions import (
    AggregationApplication,
    Conjunction,
    Implication,
    Union,
)
from .constraints_representation import RightImplication, reachable_code


class AdornedSymbol(Symbol):
    def __init__(self, expression, adornment, number):
        self.expression = expression
        self.adornment = adornment
        self.number = number

    @property
    def name(self):
        if isinstance(self.expression, Symbol):
            return self.expression.name
        else:
            raise NotImplementedError()

    def __eq__(self, other):
        return (
            hash(self) == hash(other)
            and isinstance(other, AdornedSymbol)
            and self.unapply() == other.unapply()
        )

    def __hash__(self):
        return hash((self.expression, self.adornment, self.number))

    def __str__(self) -> str:
        if isinstance(self.expression, Symbol):
            rep = self.expression.name
        elif isinstance(self.expression, Constant):
            rep = self.expression.value

        if len(self.adornment) > 0:
            superindex = f'^{self.adornment}'
        else:
            superindex = ''

        if self.number is not None:
            subindex = f'_{self.number}'
        else:
            subindex = ''
        return f'{rep}{superindex}{subindex}'

    def __repr__(self):
        if config.expression_type_printing():
            return (
                f'S{{{self}: '
                f'{self.__type_repr__}}}'
            )
        else:
            return f'S{{{self}}}'


class ReplaceAdornedSymbolWalker(ExpressionWalker):
    @add_match(Symbol)
    def replace_adorned_symbol(self, symbol):
        if isinstance(symbol, AdornedSymbol) and isinstance(
            symbol.expression, Symbol
        ):
            superindex = (
                f"^{symbol.adornment}" if len(symbol.adornment) > 0 else ""
            )
            subindex = f"_{symbol.number}" if symbol.number is not None else ""
            new_name = f"{symbol.name}{superindex}{subindex}"
            return Symbol(new_name)
        else:
            return symbol

    @add_match(Implication)
    def implication(self, exp):
        return Implication(
            self.walk(exp.consequent), self.walk(exp.antecedent)
        )


class SIPS(ABC):
    """
    Sideways Information Passing Strategy (SIPS). A SIPS defines how
    bound variables are passed from the head of a rule to the body
    predicates of the rule. SIPS formally describe what information
    (bound variables) is passed by one literal, or a conjunction of
    literals, to another literal.
    """

    def __init__(self, datalog) -> None:
        self.datalog = datalog
        self.edb = edb_with_prob_symbols(datalog)
        try:
            self.prob_symbols = datalog.probabilistic_predicate_symbols
        except AttributeError:
            self.prob_symbols = set()

    def _check_for_bound_aggregates(self, rule, adornment):
        """
        Check if there are aggregate functions in the consequent's arguments
        and raise an exception if these arguments are bound in the adornment.
        """
        bound_aggregates = {
            arg
            for arg, ad in zip(rule.consequent.args, adornment)
            if isinstance(arg, AggregationApplication) and ad == "b"
        }
        if len(bound_aggregates) > 0:
            bound_aggregate = AdornedSymbol(
                rule.consequent.functor, adornment, None
            )
            bound_aggregate = bound_aggregate(*rule.consequent.args)
            raise BoundAggregationApplicationError(
                "Magic Sets rewrite would lead to aggregation application"
                " being bound. Problematic adorned expression is: "
                f"{bound_aggregate}"
            )

    def creates_arcs(self, rule, adorned_head):
        """
        For a given rule and an adorned head predicate, create the
        arcs corresponding to this SIPS.

        For each predicate in the rule's antecedent, this method will try to
        create an arc by calling `self._create_arc`, which should return None
        if no arc can be created for this predicate, or return the head and
        tail of the arc to be added.

        Returns
        -------
        Union[Dict[Expression, Tuple[Expression]], Tuple[Expression]]
            returns the created arcs as a dict mapping of head -> tails and
            the adorned antecedent of the rule
        """
        self._check_for_bound_aggregates(rule, adorned_head.functor.adornment)

        tail_predicates = [adorned_head]
        bound_variables = {
            arg
            for arg, ad in zip(
                adorned_head.args, adorned_head.functor.adornment
            )
            if isinstance(arg, Symbol) and ad == "b"
        }
        arcs = dict()
        adorned_antecedent = ()
        checked_predicates = {}

        predicates = extract_logic_predicates(rule.antecedent)
        for predicate in predicates:
            predicate_number = checked_predicates.get(predicate, 0)
            checked_predicates[predicate] = predicate_number + 1
            arc = self._create_arc(
                    predicate,
                    predicate_number,
                    bound_variables,
                    tail_predicates,
                )
            if arc is not None:
                tail, head = arc
                arcs[head] = tail
                adorned_antecedent += (head,)
            else:
                adorned_antecedent += (predicate,)
        return arcs, adorned_antecedent

    def _adorn_predicate(
        self, predicate, predicate_number, bound_variables
    ):
        adornment = ""
        has_b = False
        for arg in predicate.args:
            if isinstance(arg, Constant) or arg in bound_variables:
                adornment += "b"
                has_b = True
            else:
                adornment += "f"

        if not has_b:
            adornment = ""

        p = AdornedSymbol(predicate.functor, adornment, predicate_number)
        p = p(*predicate.args)
        return p

    @abstractmethod
    def _create_arc(
        self,
        predicate: Expression,
        predicate_number: int,
        bound_variables: Set[Symbol],
        tail_predicates: List[Expression],
    ) -> Union[Tuple[Expression], Expression, None]:
        """
        Create an arc for the given predicate. An arc is a tuple (tail, head)
        where tail is a tuple of adorned expressions, and head is the input
        predicate adorned with the bound variables for this arc.
        This method should return None if no arc should be created for the
        given predicate.
        """
        pass


class LeftToRightSIPS(SIPS):
    """
    LeftToRightSIPS which corresponds to the default SIPS as specified in
    Balbin et al. [1].

    For a given body predicate P and head predicate H with adornment a,
    a variable v of P is bound iif:
        - it corresponds to a bound variable of H in a.
        - or it is a variable of a positive non probabilistic body literal
        left of P in the rule.

    .. [1] Isaac Balbin, Graeme S. Port, Kotagiri Ramamohanarao,
       Krishnamurthy Meenakshi. 1991. Efficient Bottom-UP Computation of
       Queries on Stratified Databases. J. Log. Program. 11(3&4). p. 305.
    """

    def __init__(self, datalog) -> None:
        super().__init__(datalog)

    def _create_arc(
        self,
        predicate: Expression,
        predicate_number: int,
        bound_variables: Set[Symbol],
        tail_predicates: List[Expression],
    ) -> Union[Tuple[Expression], Expression]:
        """
        For each arc, the tail of the arc is composed of all the predicates
        that have been adorned before it in the rule.
        This is achieved by updating the list of tail_predicates with each new
        adorned predicate.
        """
        # 1. unwrap negative predicates
        is_neg = isinstance(predicate, Negation)
        if is_neg:
            pred = predicate.formula
            if not hasattr(pred, "functor"):
                raise NegationInMagicSetsRewriteError(
                    "MagicSets can only handle negations on single predicates."
                    f"Negation {predicate} cannot be handled"
                )
        else:
            pred = predicate
        # 2. Constants, constraints and predicates in the EDB are already bound
        # and should not be adorned.
        if (
            isinstance(pred.functor, Constant) or
            pred.functor.name in self.edb or
            (
                hasattr(self.datalog, 'categorized_constraints') and
                pred.functor in self.datalog.categorized_constraints
            )
        ):
            return None

        # 3. Adorn the predicate
        p = self._adorn_predicate(pred, predicate_number, bound_variables)
        tail = tuple(tail_predicates)

        # 4. Update the tail_predicates and list of bound_variables with the
        # variables of this predicate. We only add bound variables for
        # positive, non probabilistic predicates
        if p.functor.name not in self.prob_symbols and not is_neg:
            bound_variables.update(
                arg for arg in predicate.args if isinstance(arg, Symbol)
                and arg in [a for t in tail for a in t.args]
            )
            tail_predicates.append(p)

        if is_neg:
            p = Negation(p)
        return tail, p


def magic_rewrite(
    query: Expression, datalog: DatalogProgram
) -> Tuple[Symbol, Union]:
    """
    Implementation of the Magic Sets method of optimization for datalog
    programs using Balbin et al [1] algorithm.

    .. [1] Isaac Balbin, Graeme S. Port, Kotagiri Ramamohanarao,
       Krishnamurthy Meenakshi. 1991. Efficient Bottom-UP Computation of
       Queries on Stratified Databases. J. Log. Program. 11(3&4). p. 311.

    query : Expression
        the head symbol of the query rule in the program
    datalog : DatalogProgram
        a processed datalog program to optimize

    Returns
    -------
    Tuple[Symbol, Union]
        the rewritten query symbol and program
    """
    # 1. Create a SIPS and use it to adorn the code
    sips = LeftToRightSIPS(datalog)
    adorned_code, constant_predicates = reachable_adorned_code(
        query, datalog, sips
    )
    # 2. Check that there is a constant in the code
    if len(constant_predicates) == 0:
        # No constants present in the code, magic sets is not usefull
        raise NoConstantPredicateFoundError(
            "No predicate with constant argument found."
        )
    # assume that the query rule is the last
    adorned_query = adorned_code.formulas[-1]

    # 3. Create magic rules
    magic_rules = create_balbin_magic_rules(
        adorned_code.formulas[:-1], sips
    )
    magic_inits = create_magic_query_inits(constant_predicates)

    # 4. Unadorn the magic rules + query to replace AdornedSymbols by
    # regular symbols (makes it easier to check symbol equality later on)
    unadorned_code = ReplaceAdornedSymbolWalker().walk(
        tuple(magic_inits) + tuple(magic_rules) + (adorned_query,)
    )
    # Ordered of the rules should be preserved, so the unadorned query should
    # be the last of the unadorned_code
    goal = unadorned_code[-1].consequent.functor
    return goal, Union(unadorned_code)


def create_magic_query_inits(constant_predicates: Iterable[AdornedSymbol]):
    """
    Create magic initialization predicates from the set of adorned predicates
    with at least one argument constant, according to Balbin et al.'s magic
    set algorithm.
    For each adorned predicate p^a(t) in the set, return an
    initialization rule of the form magic_p^a(t_d) :- True, where t_d is the
    vector of arguments which are bound in the adornment a of p.

    Parameters
    ----------
    constant_predicates : Iterable[AdornedExpression]
        the set of adorned predicates where at least one argument is a constant

    Returns
    -------
    Iterable[Expression]
        the set of magic initialization rules
    """
    magic_init_rules = []
    for predicate in constant_predicates:
        magic_init_rules.append(
            Implication(
                magic_predicate(predicate, adorned=False),
                Constant(True),
            )
        )
    return magic_init_rules


def create_balbin_magic_rules(adorned_rules, sips):
    """
    Create the set of Magic Set rules according to Algorithm 2 of
    Balbin et al.

    This method creates the ensemble Pm of rewritten rules from the ensemble
    P^a of adorned rules in the program (not including the adorned query rule).

    The pseudo-code algorithm for this method is:

    ```
    for each adorned rule Ra in P^a of the form head :- body
        add the rule head :- Magic(head), body to Pm
        for each adorned predicate p in body :
            add the rule Magic(p) :- Magic(h), ^(body predicates left of p)
    ```
    """
    magic_rules = []
    for rule in adorned_rules:
        consequent = rule.consequent
        magic_head = magic_predicate(consequent, adorned=False)
        if len(magic_head.args) == 0:
            magic_rules.append(rule)
            continue

        arcs, adorned_antecedent = sips.creates_arcs(rule, consequent)

        # Add the rule head :- magic(head), body
        magic_rules.append(
            Implication(
                rule.consequent,
                Conjunction((magic_head,) + adorned_antecedent),
            )
        )
        # Add the magic rules for each arc of the sips
        for head, tail in arcs.items():
            if isinstance(head, Negation):
                head = head.formula
            if "b" in head.functor.adornment:
                new_predicate = magic_predicate(head, adorned=False)
                body_lits = [
                    p if p != consequent else magic_head for p in tail
                ]
                if not (len(body_lits) == 1 and new_predicate == body_lits[0]):
                    # avoid adding rules of the form magic_p(x) :- magic_p(x)
                    magic_rules.append(
                        Implication(new_predicate, Conjunction(body_lits))
                    )

    return magic_rules


def magic_predicate(predicate, i=None, adorned=True):
    """
    Given a predicate of the form P^bf(x, y), create a magic predicate of the
    form magic_P^bf(x).
    If adorned is True, the magic predicate returned is an AdornedSymbol,
    otherwise it is a Symbol with the name magic_P^bf.

    Parameters
    ----------
    predicate : Expression
        a predicate with an adorned symbol
    i : int, optional
        a rule number, by default None
    adorned : bool, optional
        whether to return an AdornedSymbol, by default True

    Returns
    -------
    Expression
        a new predicate.
    """
    name = predicate.functor.name
    adornment = predicate.functor.adornment
    if i is not None:
        new_name = f"magic_r{i}_{name}"
    else:
        new_name = f"magic_{name}"

    new_args = [arg for arg, ad in zip(predicate.args, adornment) if ad == "b"]
    if adorned:
        new_functor = AdornedSymbol(
            Symbol(new_name), adornment, predicate.functor.number
        )
    else:
        superindex = f"^{adornment}" if len(adornment) > 0 else ""
        subindex = (
            f"_{predicate.functor.number}"
            if predicate.functor.number is not None
            else ""
        )
        new_name = f"{new_name}{superindex}{subindex}"
        new_functor = Symbol(new_name)
    return new_functor(*new_args)


def reachable_adorned_code(query, datalog, sips: SIPS):
    adorned_code, constant_predicates = adorn_code(query, datalog, sips)
    adorned_datalog = type(datalog)()
    adorned_datalog.walk(adorned_code)
    # assume that the query rule is the first
    adorned_query = adorned_code.formulas[0]
    return (
        reachable_code(adorned_query, adorned_datalog),
        constant_predicates,
    )


def adorn_code(
    query: Expression, datalog: DatalogProgram, sips: SIPS
) -> Union:
    """
    Produce the rewritten datalog program according to the
    Magic Sets technique.

    Parameters
    ----------
    query : Expression
        query to solve
    datalog : DatalogProgram
        processed datalog program
    sips_class : SIPS
        the SIPS instance to use for adornment of predicates

    Returns
    -------
    Union
        adorned code where the query rule is the first expression
        in the block.
    """
    adornment = ''
    for a in query.args:
        if isinstance(a, Symbol) or isinstance(a, ProbabilisticQuery):
            adornment += 'f'
        else:
            adornment += 'b'

    query = AdornedSymbol(query.functor, adornment, 0)(*query.args)
    adorn_stack = [query]

    idb = datalog.intensional_database()
    rewritten_program = []
    rewritten_rules = set()
    constant_predicates = set()

    while adorn_stack:
        consequent = adorn_stack.pop()

        if isinstance(consequent.functor, AdornedSymbol):
            adornment = consequent.functor.adornment
            name = consequent.functor.expression.name
        else:
            adornment = ''
            name = consequent.functor.name

        rules = idb.get(name, None)
        if rules is None:
            continue

        for rule in rules.formulas:
            if isinstance(rule, RightImplication):
                continue
            new_consequent = consequent.functor(*rule.consequent.args)
            adorned_antecedent, to_adorn = adorn_antecedent(
                rule, new_consequent, rewritten_rules, sips
            )
            adorn_stack += to_adorn
            rewritten_program.append(
                Implication(new_consequent, adorned_antecedent)
            )
            rewritten_rules.add(consequent.functor)
            for predicate in to_adorn:
                if any([isinstance(arg, Constant) for arg in predicate.args]):
                    constant_predicates.add(predicate)

    return Union(rewritten_program), constant_predicates


def adorn_antecedent(
    rule, adorned_head, rewritten_rules, sips: SIPS
):
    to_adorn = []
    arcs, adorned_antecedent = sips.creates_arcs(rule, adorned_head)
    for adorned_predicate in arcs.keys():
        if isinstance(adorned_predicate, Negation):
            adorned_predicate = adorned_predicate.formula
        if adorned_predicate.functor not in rewritten_rules:
            to_adorn.append(adorned_predicate)

    antecedent = rule.antecedent

    if len(adorned_antecedent) == 1:
        adorned_antecedent = adorned_antecedent[0]
    elif not isinstance(antecedent, Conjunction):
        raise NonConjunctiveAntecedentInMagicSetsError(
            "Magic Set rewrite does not work with "
            f"non-conjunctive antecedent {antecedent}"
        )
    else:
        adorned_antecedent = Conjunction(adorned_antecedent)
    return adorned_antecedent, to_adorn


def edb_with_prob_symbols(datalog: DatalogProgram) -> Iterable[Symbol]:
    edb = set(datalog.extensional_database()) | set(datalog.included_functions)
    try:
        edb = (
            edb
            | set(datalog.probabilistic_facts())
            | set(datalog.probabilistic_choices())
        )
    except AttributeError:
        pass
    return edb



from inspect import signature
from itertools import product
from operator import add, eq, mul, pow, sub, truediv
import logging

import pytest

from .... import config
from ....datalog import Conjunction, Fact, Implication, Negation, Union
from ....datalog.aggregation import AggregationApplication
from ....expression_pattern_matching import add_match
from ....expression_walker import ExpressionWalker, ReplaceExpressionWalker
from ....expressions import Constant, Query, Symbol
from ....logic import (
    ExistentialPredicate,
    LogicOperator,
    NaryLogicOperator,
    UniversalPredicate
)
from ....probabilistic.expressions import ProbabilisticPredicate
from ..squall import LogicSimplifier
from ..squall_syntax_lark import parser
from ..standard_syntax import ExternalSymbol
from ...probabilistic_frontend import RegionFrontendCPLogicSolver, Chase


LOGGER = logging.getLogger()


config.disable_expression_type_printing()


EQ = Constant(eq)


class LogicWeakEquivalence(ExpressionWalker):
    @add_match(EQ(UniversalPredicate, UniversalPredicate))
    def eq_universal_predicates(self, expression):
        left, right = expression.args
        if left.head != right.head:
            new_head = Symbol.fresh()
            rew = ReplaceExpressionWalker(
                {left.head: new_head, right.head: new_head}
            )
            left = rew.walk(left)
            right = rew.walk(right)

        return self.walk(EQ(left.body, right.body))

    @add_match(EQ(ExistentialPredicate, ExistentialPredicate))
    def eq_existential_predicates(self, expression):
        return self.eq_universal_predicates(expression)

    @add_match(
        EQ(NaryLogicOperator, NaryLogicOperator),
        lambda exp: (
            len(exp.args[0].formulas) == len(exp.args[1].formulas) and
            type(exp.args[0]) == type(exp.args[1])
        )
    )
    def n_ary_sort(self, expression):
        left, right = expression.args

        return all(
            self.walk(EQ(a1, a2))
            for a1, a2 in zip(
                sorted(left.formulas, key=repr),
                sorted(right.formulas, key=repr)
            )
        )

    @add_match(EQ(LogicOperator, LogicOperator))
    def eq_logic_operator(self, expression):
        left, right = expression.args
        for l, r in zip(left.unapply(), right.unapply()):
            if isinstance(l, tuple) and isinstance(r, tuple):
                return (
                    (len(l) == len(r)) and
                    all(
                        self.walk(EQ(ll, rr))
                        for ll, rr in zip(l, r)
                    )
                )
            else:
                return self.walk(EQ(l, r))

    @add_match(EQ(..., ...))
    def eq_expression(self, expression):
        return expression.args[0] == expression.args[1]


def weak_logic_eq(left, right):
    left = LogicSimplifier().walk(left)
    right = LogicSimplifier().walk(right)
    return LogicWeakEquivalence().walk(EQ(left, right))


@pytest.fixture(scope="module")
def nouns():
    return [
        ("'marseillaise'", lambda x: x(Constant('marseillaise'))),
        ("?x", lambda x:x(Symbol('x')))
    ]


@pytest.fixture(scope="module")
def verb1():
    return [
        ("plays", lambda x: Symbol("plays")(x)),
    ]


@pytest.fixture(scope="module")
def verb2():
    return [
        ("~sings", lambda x, y: Symbol("sings")(x, y))
    ]


@pytest.fixture(scope="module")
def verbs(verb1, verb2):
    return verb1 + verb2


def op_application(np):
    return lambda d: np(lambda y: d(y))


@pytest.fixture(scope="module")
def op(noun_phrases):
    return [
        (np[0], op_application(np[1]))
        for np in noun_phrases
    ]


def vp_op_application(op, v2):
    return lambda x: op(lambda y: v2(x, y))


@pytest.fixture(scope="module")
def verb_phrases_do_op(verbs, op):
    return [
        (f"{v2[0]} {op[0]}", vp_op_application(op[1], v2[1]))
        for v2, op in product(verbs, op)
        if is_transitive(v2[1])
    ]


@pytest.fixture(scope="module")
def nouns_1():
    return [
        ("person", Symbol("person")),
        ("country", Symbol("country"))
    ]


@pytest.fixture(scope="module")
def nouns_2():
    return [
        ("~author", Symbol("author")),
        ("~publication_year", Symbol("publication_year"))
    ]


def lambda_simple(arg):
    return lambda x: arg(x)


def lambda_conjunction(*args):
    return lambda x: Conjunction(tuple(a(x) for a in args))


def lambda_conjunction_2(arg2, *args):
    return lambda x: lambda y: Conjunction(
        tuple(a(x) for a in args) + (arg2(x, y),)
    )


@pytest.fixture(scope="module")
def noun_groups_1(nouns_1):
    return [
        (f"{n1[0]}", lambda_conjunction(n1[1]))
        for n1 in nouns_1
    ]


@pytest.fixture(scope="module")
def noun_groups_2(nouns_2):
    return [
        (f"{n2[0]}", lambda_conjunction_2(n2[1]))
        for n2 in nouns_2
    ]



@pytest.fixture(scope="module")
def det():
    res = []

    fresh = lambda: Symbol.fresh()
    res.append((
        "every",
        lambda d1: lambda d2: (
            (
                lambda x: UniversalPredicate(x, Implication(d2(x), d1(x))))
                (fresh()
            )
        )
    ))

    for det in ("a", "an", "some"):
        res.append((
            det,
            lambda d1: lambda d2: (
                (lambda x: ExistentialPredicate(x, Conjunction((d1(x), d2(x)))))(fresh())
            )
        ))

    res.append((
        "no",
        lambda d1: lambda d2: (
            (lambda x: Negation(ExistentialPredicate(x, Conjunction((d1(x), d2(x))))))(fresh())
        )
    ))
    return res


@pytest.fixture(scope="module")
def noun_phrase_quantified_1(det, noun_groups_1):
    res = []
    for det_, ng1 in product(det, noun_groups_1):
        exp = det_[1](ng1[1])
        res.append((
            f"{det_[0]} {ng1[0]}", lambda_simple(exp)
        ))
    return res


def np_np2_composition(np, np2):
    return lambda d: np(lambda x: (np2(x)(d)))


@pytest.fixture(scope="module")
def noun_phrase_quantified_2(noun_phrase_2, nouns):
    res = []
    for np2, np in product(noun_phrase_2, nouns):
        exp = np_np2_composition(np[1], np2[1])
        res.append((f"{np2[0]} of {np[0]}", exp))
    return res


def det_ng2_composition(det, ng2):
    return lambda x: lambda d: (
        det(lambda y: ng2(x)(y))(d)
    )


@pytest.fixture(scope="module")
def noun_phrase_2(det, noun_groups_2):
    res = []
    for det_, ng2 in product(det, noun_groups_2):
        exp = det_ng2_composition(det_[1], ng2[1])
        res.append((
            f"{det_[0]} {ng2[0]}", exp
        ))
    return res


@pytest.fixture(scope="module")
def verb_phrases(verbs, verb_phrases_do_op):
    return [
        v for v in verbs if not is_transitive(v[1])
    ] + verb_phrases_do_op


@pytest.fixture(scope="module")
def noun_phrases(nouns, noun_phrase_quantified_1, noun_phrase_quantified_2):
    return nouns + noun_phrase_quantified_1 + noun_phrase_quantified_2 


@pytest.fixture(scope="module")
def s_np_vp(noun_phrases, verb_phrases):
    return [
        (f"{np[0]} {vp[0]}", np[1](vp[1]))
        for np, vp in product(noun_phrases, verb_phrases)
    ]


@pytest.fixture(scope="module")
def s_for(s_np_vp, noun_phrases):
    res = list(s_np_vp)
    for _ in range(1):
        res += [
            (f"for {np[0]}, {s[0]}", np[1](lambda x: s[1]))
            for np, s in product(noun_phrases, res)
        ]
    res = res[len(s_np_vp):]
    return res


@pytest.fixture(scope="module")
def s(s_np_vp, s_for):
    return s_np_vp + s_for


@pytest.mark.slow
def test_squall_s(s):
    for query, expected in s:
        LOGGER.log(logging.INFO + 1, "Query to test %s", query)
        res = parser(f"squall {query}")
        assert weak_logic_eq(res, expected)


def is_transitive(verb):
    return len(signature(verb).parameters) > 1


def test_squall_simple_np_nv(noun_phrases, verb_phrases):
    query_result_pairs = []
    for np, vp in product(noun_phrases, verb_phrases):
        query = f"{np[0]} {vp[0]}"
        result = np[1](vp[1])
        query_result_pairs.append((query, result))

    for query, expected in query_result_pairs:
        res = parser(f"squall {query}")
        assert weak_logic_eq(res, expected)


def test_squall_quantified_np_nv(noun_phrase_quantified_1, verb_phrases):
    query_result_pairs = []
    for np, vp in product(noun_phrase_quantified_1, verb_phrases):
        query_result_pairs.append(
            (f"{np[0]} {vp[0]}", np[1](vp[1]))
        )

    for query, expected in query_result_pairs:
        res = parser(f"squall {query}")
        assert weak_logic_eq(res, expected)


def test_squall_quantified_np2_nv(noun_phrase_quantified_2, verb_phrases):
    query_result_pairs = []
    for np, vp in product(noun_phrase_quantified_2, verb_phrases):
        query_result_pairs.append(
            (f"{np[0]} {vp[0]}", np[1](vp[1]))
        )

    for query, expected in query_result_pairs:
        res = parser(f"squall {query}")
        assert weak_logic_eq(res, expected)


def test_squall_quantified_np_nv2(
    noun_phrase_quantified_1, verb2, noun_phrases
):
    query_result_pairs = []
    for npq, vp, np in product(
        noun_phrase_quantified_1, verb2, noun_phrases
    ):
        vp_ = lambda x: np[1](lambda y: vp[1](x, y))
        query_result_pairs.append(
            (f"{npq[0]} {vp[0]} {np[0]}", npq[1](vp_))
        )

    for query, expected in query_result_pairs:
        res = parser(f"squall {query}")
        assert weak_logic_eq(res, expected)


def test_squall_voxel_activation():
    query = "every voxel (?x; ?y; ?z) that a study ?s ~reports activates"
    res = parser(f"squall {query}")
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    s = Symbol("s")
    voxel = Symbol("voxel")
    study = Symbol("study")
    reports = Symbol("reports")
    activates = Symbol("activates")
    expected = UniversalPredicate(
        z,
        UniversalPredicate(
            y,
            UniversalPredicate(
                x,
                Implication(
                    activates(x, y, z),
                    Conjunction((
                        voxel(x, y, z),
                        ExistentialPredicate(
                            s,
                            Conjunction((
                                reports(s, x, y, z),
                                study(s)
                            ))
                        )

                    ))
                )
            )
        )
    )

    assert weak_logic_eq(res, expected)


def test_squall():
    # res = parser("squall ?t1 runs")
    # res = parser("squall every woman sees a man")
    # res = parser("squall a woman that eats sees a man that sleeps")
    res = parser("squall ?s reports")
    parser("squall every voxel that a study ~reports is an activation")
    parser("squall every voxel that 'neuro study' ~reports activates")
    parser("squall every voxel that a study that ~mentions a word that is 'auditory' ~reports, activates")
    parser("squall every voxel that a study that ~reports no region ~reports activates")
    # parser("squall every voxel that a study --that ~reports no region-- ~reports activates")

    assert res


@pytest.fixture
def datalog_simple():
    datalog = RegionFrontendCPLogicSolver()

    datalog.add_extensional_predicate_from_tuples(
        Symbol("item"),
        [('a',), ('b',), ('c',), ('d',)]
    )

    datalog.add_extensional_predicate_from_tuples(
        Symbol("item_count"),
        [('a', 0), ('a', 1), ('b', 2), ('c', 3)]
    )

    datalog.add_extensional_predicate_from_tuples(
        Symbol("quantity"),
        [(i,) for i in range(5)]
    )
    return datalog


def test_lark_semantics_item_selection(datalog_simple):
    code = (
        "define as Large every Item "
        "that has an item_count greater equal than 2."
    )
    logic_code = parser(code)
    datalog_simple.walk(logic_code)
    chase = Chase(datalog_simple)
    solution = chase.build_chase_solution()['large'].value
    expected = set([('b',), ('c',)])

    assert solution == expected


def test_lark_semantics_join(datalog_simple):
    code = (
        "define as merge for every Item ?i ;"
        " with every Quantity that ?i item_counts"
    )
    logic_code = parser(code)
    datalog_simple.walk(logic_code)
    chase = Chase(datalog_simple)
    solution = chase.build_chase_solution()['merge'].value
    expected = set([('a', 0), ('a', 1), ('b', 2), ('c', 3)])

    assert solution == expected


def test_lark_semantics_aggregation(datalog_simple):
    code = """
        define as max_items for every Item ?i ;
            where every Max of the Quantity where ?i item_count per ?i.
    """
    logic_code = parser(code)

    datalog_simple.walk(logic_code)
    chase = Chase(datalog_simple)
    solution = chase.build_chase_solution()["max_item"].value
    expected = set([('a', 1), ('b', 2), ('c', 3)])
    assert solution == expected



import operator
from itertools import chain

from .. import expression_walker as ew
from ..expressions import Constant, FunctionApplication, Symbol
from .relational_algebra import (
    COUNT,
    Column,
    ColumnInt,
    ColumnStr,
    EquiJoin,
    ExtendedProjection,
    FullOuterNaturalJoin,
    FunctionApplicationListMember,
    GroupByAggregation,
    LeftNaturalJoin,
    NameColumns,
    NaturalJoin,
    NumberColumns,
    Product,
    Projection,
    RelationalAlgebraOperation,
    RelationalAlgebraStringExpression,
    RenameColumn,
    RenameColumns,
    ReplaceNull,
    Selection,
    eq_,
    get_expression_columns,
    int2columnint_constant,
    str2columnstr_constant
)

AND = Constant(operator.and_)
EQ = Constant(operator.eq)


class ProductSimplification(ew.ExpressionWalker):
    @ew.add_match(Product, lambda x: len(x.relations) == 1)
    def single_product(self, product):
        return self.walk(product.relations[0])


class RewriteSelections(ew.ExpressionWalker):
    """
    Mixing that optimises through relational algebra expressions.

    The pushing selections (:obj:`Selection`) down and compositions
    and reduces the reach of selections. Then it converts
    equi-selection/product compositions into equijoins.
    """

    @ew.add_match(
        Selection(..., FunctionApplication(eq_, (..., Constant[Column]))),
        lambda s: not issubclass(s.formula.args[0].type, Column),
    )
    def swap_formula_args(self, selection):
        new_selection = Selection(
            self.walk(selection.relation), eq_(*selection.formula.args[::-1])
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            ..., FunctionApplication(eq_, (Constant[Column], Constant[Column]))
        ),
        lambda s: s.formula.args[0].value > s.formula.args[1].value,
    )
    def sort_formula_args(self, selection):
        new_selection = Selection(
            self.walk(selection.relation), eq_(*selection.formula.args[::-1])
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            Selection(..., FunctionApplication(eq_, (Constant[Column], ...))),
            FunctionApplication(eq_, (Constant[Column], ...)),
        ),
        lambda s: s.formula.args[0].value > s.relation.formula.args[0].value,
    )
    def selection_selection_swap(self, selection):
        new_selection = Selection(
            Selection(selection.relation.relation, selection.formula),
            selection.relation.formula,
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        ),
        lambda s: (
            s.formula.args[0].value
            >= RewriteSelections.get_arity(
                s.relation.relations[0]
            )
        ),
    )
    def selection_push_right(self, selection):
        relations = selection.relation.relations
        column = int(selection.formula.args[0].value)

        i, accum_arity, column = self.split_relations_column(relations, column)

        left_relations = relations[:i]
        relations = relations[i:]

        arg_right = selection.formula.args[1]
        if issubclass(arg_right.type, Column):
            arg_right = Constant[ColumnInt](
                ColumnInt(-accum_arity + arg_right.value)
            )

        res = Product(
            left_relations
            + (
                Selection(
                    Product(relations),
                    eq_(Constant[ColumnInt](ColumnInt(column)), arg_right),
                ),
            )
        )
        return self.walk(res)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        ),
        lambda s: (
            s.formula.args[1].value
            < (
                RewriteSelections.get_arity(s.relation)
                - RewriteSelections.get_arity(
                    s.relation.relations[-1]
                )
            )
        ),
    )
    def selection_shorten_right(self, selection):
        relations = selection.relation.relations
        column = int(selection.formula.args[1].value)

        i, accum_arity, column = self.split_relations_column(relations, column)
        column += accum_arity
        inner_relations = tuple(relations[: i + 1])
        if len(inner_relations) == 1:
            inner_relations = inner_relations[0]
        else:
            inner_relations = Product(inner_relations)

        outer_relations = tuple(relations[i + 1:])

        arg_left = selection.formula.args[0]

        res = Product(
            (
                Selection(
                    inner_relations,
                    eq_(arg_left, Constant[ColumnInt](ColumnInt(column))),
                ),
            )
            + outer_relations
        )
        return self.walk(res)

    @ew.add_match(
        Selection(
            EquiJoin,
            FunctionApplication(
                eq_, (Constant[ColumnInt], Constant[ColumnInt])
            ),
        )
    )
    def selection_on_equijoin_columns(self, selection):
        column_left = selection.formula.args[0].value
        column_right = selection.formula.args[1].value
        column_min, column_max = sorted((column_left, column_right))
        relation_left = selection.relation.relation_left
        relation_right = selection.relation.relation_right
        left_arity = self.get_arity(relation_left)
        if column_max < left_arity:
            relation_left = Selection(relation_left, selection.formula)
        elif column_min >= left_arity:
            column_min -= left_arity
            column_max -= left_arity
            new_formula = eq_(
                Constant[ColumnInt](ColumnInt(column_min)),
                Constant[ColumnInt](ColumnInt(column_max)),
            )
            relation_right = Selection(relation_right, new_formula)
        else:
            return selection

        return self.walk(
            EquiJoin(
                relation_left,
                selection.relation.columns_left,
                relation_right,
                selection.relation.columns_right,
            )
        )

    @ew.add_match(
        Selection(
            EquiJoin, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        )
    )
    def selection_on_equijoin(self, selection):
        column = selection.formula.args[0].value
        relation_left = selection.relation.relation_left
        relation_right = selection.relation.relation_right
        left_arity = self.get_arity(relation_left)
        if column < left_arity:
            relation_left = Selection(relation_left, selection.formula)
        else:
            relation_right = Selection(
                relation_right,
                eq_(
                    Constant(
                        ColumnInt(
                            int(selection.formula.args[0].value) - left_arity
                        )
                    ),
                    selection.formula.args[1],
                ),
            )

        return self.walk(
            EquiJoin(
                relation_left,
                selection.relation.columns_left,
                relation_right,
                selection.relation.columns_right,
            )
        )

    @ew.add_match(
        Selection(
            Product,
            FunctionApplication(
                eq_, (Constant[ColumnInt], Constant[ColumnInt])
            ),
        )
    )
    def selection_between_columns_product(self, selection):
        relations = selection.relation.relations
        if len(relations) == 1:
            res = Selection(relations[0], selection.formula)
        else:
            column_left = (selection.formula.args[0],)

            column_right = int(selection.formula.args[1].value)
            left_arity = self.get_arity(relations[0])
            column_right -= left_arity
            column_right = (Constant[ColumnInt](ColumnInt(column_right)),)

            relations_right = relations[1:]
            if len(relations_right) == 1:
                relation_right = relations_right[0]
            else:
                relation_right = Product(relations_right)

            res = EquiJoin(
                relations[0], column_left, relation_right, column_right
            )

        return self.walk(res)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        )
    )
    def selection_by_constant_on_product(self, selection):
        return self.walk(
            Product(
                (
                    Selection(
                        selection.relation.relations[0], selection.formula
                    ),
                )
                + selection.relation.relations[1:]
            )
        )

    @ew.add_match(
       Selection(Selection, ...),
       lambda exp: all(
           isinstance(col, Constant[ColumnStr])
           for col in (
               get_expression_columns(exp.formula) |
               get_expression_columns(exp.relation.formula)
           )
       )
    )
    def merge_selections(self, expression):
        return self.walk(
            Selection(
                expression.relation.relation,
                AND(expression.relation.formula, expression.formula)
            )
        )

    @staticmethod
    def split_relations_column(relations, column):
        accum_arity = 0
        for i, relation in enumerate(relations):
            current_arity = RewriteSelections.get_arity(
                relation
            )
            if column < current_arity:
                break
            accum_arity += current_arity
            column -= current_arity
        return i, accum_arity, column

    @staticmethod
    def get_arity(expression):
        if isinstance(expression, Constant):
            return expression.value.arity
        elif isinstance(expression, Product):
            return sum(
                RewriteSelections.get_arity(r)
                for r in expression.relations
            )
        elif isinstance(expression, EquiJoin):
            return RewriteSelections.get_arity(
                expression.relation_left
            ) + RewriteSelections.get_arity(
                expression.relation_right
            )
        else:
            return RewriteSelections.get_arity(
                expression.relation
            )


class EliminateTrivialProjections(ew.PatternWalker):
    @ew.add_match(
        Projection(Constant, ...),
        lambda expression: (
            tuple(c.value for c in expression.attributes)
            == tuple(c for c in expression.relation.value.columns)
        ) or (
            tuple(str(c.value) for c in expression.attributes)
            == tuple(c for c in expression.relation.value.columns)
        )
    )
    def eliminate_trivial_projection(self, expression):
        return expression.relation

    @ew.add_match(
        Projection(Projection, ...),
        lambda expression: all(
            isinstance(attribute, Constant[ColumnInt])
            for attribute in (
                expression.attributes + expression.relation.attributes
            )
        )
    )
    def eliminate_trivial_nested_unnamed_projection(self, expression):
        inner_attributes = expression.relation.attributes
        outer_attributes = expression.attributes

        resulting_attributes = tuple(
            inner_attributes[attribute.value]
            for attribute in outer_attributes
        )

        return self.walk(
            Projection(expression.relation.relation, resulting_attributes)
        )

    @ew.add_match(Projection(Projection, ...))
    def eliminate_trivial_nested_projection(self, expression):
        return self.walk(
            Projection(expression.relation.relation, expression.attributes)
        )

    @ew.add_match(Projection(ExtendedProjection, ...))
    def simplify_projection_extended_projection(self, expression):
        new_projections = tuple(
            proj for proj in expression.relation.projection_list
            if proj.dst_column in expression.attributes
        )
        new_relation = self.walk(
            ExtendedProjection(
                expression.relation.relation,
                new_projections
            )
        )
        return new_relation

    @ew.add_match(
        ExtendedProjection,
        lambda e: all(
            isinstance(p.fun_exp, Constant[ColumnStr]) and
            (p.fun_exp == p.dst_column)
            for p in e.projection_list
        )
    )
    def convert_extended_projection_2_projection(self, expression):
        return self.walk(Projection(
            expression.relation,
            tuple(p.dst_column for p in expression.projection_list)
        ))

    @ew.add_match(
        ExtendedProjection,
        lambda e: all(
            isinstance(p.fun_exp, Constant[ColumnStr])
            for p in e.projection_list
        ) and (
            len(set(p.fun_exp for p in e.projection_list))
            == len(e.projection_list)
            == len(e.relation.columns())
        )
    )
    def convert_extended_projection_2_rename(self, expression):
        return self.walk(RenameColumns(
            expression.relation,
            tuple(
                (p.fun_exp, p.dst_column)
                for p in expression.projection_list
            )
        ))


class SimplifyExtendedProjectionsWithConstants(ew.PatternWalker):
    @ew.add_match(
        ExtendedProjection(
            ExtendedProjection,
            ...
        ),
        lambda expression: all(
            isinstance(int_proj.fun_exp, (Constant, FunctionApplication))
            for int_proj in expression.relation.projection_list
        )
    )
    def nested_extended_projection_constant(self, expression):
        rew = ew.ReplaceExpressionWalker({
            p.dst_column: p.fun_exp
            for p in expression.relation.projection_list
        })
        new_projections = (
            FunctionApplicationListMember(rew.walk(p.fun_exp), p.dst_column)
            for p in expression.projection_list
        )
        return self.walk(
            ExtendedProjection(
                expression.relation.relation,
                new_projections
            )
        )

    @ew.add_match(
        ExtendedProjection(
            NaturalJoin(ExtendedProjection, ...),
            ...
        ),
        lambda expression: any(
            _function_application_list_member_has_constant_exp(int_proj)
            for int_proj in expression.relation.relation_left.projection_list
        )
    )
    def nested_extended_projection_naturaljoin_constant_l(self, expression):
        return self._nested_ep_join_constant_left(expression)

    @ew.add_match(
        ExtendedProjection(
            LeftNaturalJoin(ExtendedProjection, ...),
            ...
        ),
        lambda expression: any(
           _function_application_list_member_has_constant_exp(int_proj)
           for int_proj in expression.relation.relation_left.projection_list
        )
    )
    def nested_extended_projection_leftnaturaljoin_constant(self, expression):
        return self._nested_ep_join_constant_left(expression)

    @ew.add_match(
        ExtendedProjection(
            FullOuterNaturalJoin(ExtendedProjection, ...),
            ...
        ),
        lambda expression: any(
           _function_application_list_member_has_constant_exp(int_proj)
           for int_proj in expression.relation.relation_left.projection_list
        )
    )
    def nested_extended_projection_outernaturaljoin_constant(self, expression):
        return self._nested_ep_join_constant_left(expression)

    @ew.add_match(
        ExtendedProjection(
            NaturalJoin(..., ExtendedProjection),
            ...
        ),
        lambda expression: any(
            _function_application_list_member_has_constant_exp(int_proj)
            for int_proj in expression.relation.relation_right.projection_list
        )
    )
    def nested_extended_projection_naturaljoin_constantr(self, expression):
        return self._nested_ep_join_constant_left(expression, flip=True)

    @ew.add_match(
        ExtendedProjection(
            LeftNaturalJoin(..., ExtendedProjection),
            ...
        ),
        lambda expression: any(
           _function_application_list_member_has_constant_exp(int_proj)
           for int_proj in expression.relation.relation_right.projection_list
        )
    )
    def nested_extended_projection_leftnaturaljoin_constantr(self, expression):
        return self._nested_ep_join_constant_left(expression, flip=True)

    @ew.add_match(
        ExtendedProjection(
            FullOuterNaturalJoin(..., ExtendedProjection),
            ...
        ),
        lambda expression: any(
           _function_application_list_member_has_constant_exp(int_proj)
           for int_proj in expression.relation.relation_right.projection_list
        )
    )
    def nested_extended_projection_outernaturaljoin_constantr(self, expression):
        return self._nested_ep_join_constant_left(expression, flip=True)

    def _nested_ep_join_constant_left(self, expression, flip=False):
        if flip:
            right, left = expression.relation.unapply()
        else:
            left, right = expression.relation.unapply()
        new_inner_projections = []
        replacements = {}
        selections = []
        right_columns = right.columns()
        for proj in left.projection_list:
            if (
                isinstance(proj.fun_exp, Constant) and
                not isinstance(proj.fun_exp, Constant[Column])
            ):
                replacements[proj.dst_column] = proj.fun_exp
                if proj.dst_column in right_columns:
                    selections.append(eq_(proj.dst_column, proj.fun_exp))
            else:
                new_inner_projections.append(proj)

        rew = ew.ReplaceExpressionWalker(replacements)
        new_external_projections = tuple(
            FunctionApplicationListMember(rew.walk(p.fun_exp), p.dst_column)
            for p in expression.projection_list
        )
        for selection in selections:
            right = Selection(right, selection)

        left = ExtendedProjection(left.relation, tuple(new_inner_projections))

        if flip:
            new_join = expression.relation.apply(right, left)
        else:
            new_join = expression.relation.apply(left, right)

        return self.walk(
            ExtendedProjection(
                new_join,
                new_external_projections
            )
        )

    @ew.add_match(
        NaturalJoin(ExtendedProjection, ...),
        lambda expression: (
            set(
                proj.dst_column
                for proj in expression.relation_left.projection_list
                if not isinstance(proj.fun_exp, Constant[Column])
            ) - expression.relation_right.columns()
        )
    )
    def push_computed_columns_up(self, expression, flip=False):
        if flip:
            right, left = expression.unapply()
        else:
            left, right = expression.unapply()

        right_columns = right.columns()
        projections_to_push_up = []
        projections_to_keep = []
        new_columns = {
            proj.dst_column: proj.dst_column
            for proj in left.projection_list
            if proj.dst_column == proj.fun_exp
        }
        for proj in left.projection_list:
            if (
                proj.dst_column not in right_columns and
                not isinstance(proj.fun_exp, Constant[Column])
            ):
                fun_columns = [
                    c
                    for _, c in ew.expression_iterator(proj.fun_exp)
                    if isinstance(c, Constant[Column])
                ]
                for c in fun_columns:
                    if c not in new_columns:
                        new_columns[c] = str2columnstr_constant(
                            Symbol.fresh().name
                        )

                projections_to_push_up.append(FunctionApplicationListMember(
                    ew.ReplaceExpressionWalker(new_columns).walk(proj.fun_exp),
                    proj.dst_column
                ))
                projections_to_keep += [
                    FunctionApplicationListMember(c, new_columns[c])
                    for _, c in ew.expression_iterator(proj.fun_exp)
                    if isinstance(c, Constant[Column])
                ]
            else:
                projections_to_keep.append(proj)

        if len(projections_to_keep) > 0:
            left = ExtendedProjection(
                left.relation,
                tuple(projections_to_keep)
            )
        if flip:
            new_join = expression.apply(right, left)
        else:
            new_join = expression.apply(left, right)

        res = ExtendedProjection(
            new_join,
            tuple(
                FunctionApplicationListMember(c, c)
                for c in new_join.columns()
            ) + tuple(projections_to_push_up)
        )
        return self.walk(res)

    @ew.add_match(
        LeftNaturalJoin(ExtendedProjection, ...),
        lambda expression: (
            set(
                proj.dst_column
                for proj in expression.relation_left.projection_list
                if not isinstance(proj.fun_exp, Constant[Column])
            ) - expression.relation_right.columns()
        )
    )
    def push_computed_columns_up_left(self, expression):
        return self.push_computed_columns_up(expression)

    @ew.add_match(
        FullOuterNaturalJoin(ExtendedProjection, ...),
        lambda expression: (
            set(
                proj.dst_column
                for proj in expression.relation_left.projection_list
                if not isinstance(proj.fun_exp, Constant[Column])
            ) - expression.relation_right.columns()
        )
    )
    def push_computed_columns_up_outer(self, expression):
        return self.push_computed_columns_up(expression)

    @ew.add_match(
        NaturalJoin(..., ExtendedProjection),
        lambda expression: (
            set(
                proj.dst_column
                for proj in expression.relation_right.projection_list
                if not isinstance(proj.fun_exp, Constant[Column])
            ) - expression.relation_left.columns()
        )
    )
    def push_computed_columns_up_flip(self, expression):
        return self.push_computed_columns_up(expression, True)

    @ew.add_match(
        GroupByAggregation(
            ExtendedProjection, ...,
            (
                FunctionApplicationListMember(
                    FunctionApplication(Constant(sum), ...),
                    ...
                ),
            )
        ),
        lambda expression: any(
            (
                proj.dst_column ==
                expression.aggregate_functions[0].fun_exp.args[0]
            ) &
            (proj.fun_exp == Constant[float](1))
            for proj in expression.relation.projection_list
        )
    )
    def replace_trivial_agg_groupby(self, expression):
        new_projections = tuple(
            proj for proj in expression.relation.projection_list
            if (
                proj.dst_column !=
                expression.aggregate_functions[0].fun_exp.args[0]
            )
        )
        return self.walk(
            GroupByAggregation(
                ExtendedProjection(
                    expression.relation.relation,
                    new_projections
                ),
                expression.groupby,
                (
                    FunctionApplicationListMember(
                        COUNT(),
                        expression.aggregate_functions[0].dst_column
                    ),
                )

            )
        )

    @ew.add_match(
        ReplaceNull(ExtendedProjection, ..., ...),
        lambda exp: all(
            isinstance(proj.fun_exp, Constant[ColumnStr])
            for proj in exp.relation.projection_list
            if exp.column == proj.dst_column
        )
    )
    def push_replace_null_in_ext_proj(self, expression):
        for proj in expression.relation.projection_list:
            if proj.dst_column == expression.column:
                replacement = proj.fun_exp
                break

        return self.walk(ExtendedProjection(
            ReplaceNull(
                expression.relation.relation,
                replacement,
                expression.value
            ),
            expression.relation.projection_list
        ))


def _function_application_list_member_has_constant_exp(int_proj):
    return (
        isinstance(int_proj.fun_exp, Constant) and
        not isinstance(
            int_proj.fun_exp,
            (Constant[Column], Constant[RelationalAlgebraStringExpression])
        )
    )


class RenameOptimizations(ew.PatternWalker):
    @ew.add_match(RenameColumn)
    def convert_rename_column(self, expression):
        return self.walk(RenameColumns(
            expression.relation,
            ((expression.src, expression.dst),)
        ))

    @ew.add_match(
        RenameColumns,
        lambda exp: (
            all(s == d for s, d in exp.renames)
        )
    )
    def remove_trivial_rename(self, expression):
        return self.walk(expression.relation)

    @ew.add_match(RenameColumns(RenameColumns, ...))
    def merge_nested_rename_columns(self, expression):
        outer_renames = {src: dst for src, dst in expression.renames}
        new_renames = []
        for src, dst in expression.relation.renames:
            if dst in outer_renames:
                new_renames.append(
                    (src, outer_renames[dst])
                )
                del outer_renames[dst]
            else:
                new_renames.append((src, dst))
        new_renames += [
            (src, dst)
            for src, dst in outer_renames.items()
        ]

        return self.walk(
            RenameColumns(expression.relation.relation, tuple(new_renames))
        )

    @ew.add_match(RenameColumns(ExtendedProjection, ...))
    def merge_rename_columns_extended_projection(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        projection_list = expression.relation.projection_list
        new_projection_list = []
        for falm in projection_list:
            if falm.dst_column in renames:
                falm = FunctionApplicationListMember(
                    falm.fun_exp,
                    renames[falm.dst_column]
                )
            new_projection_list.append(falm)
        return self.walk(ExtendedProjection(
            expression.relation.relation,
            tuple(new_projection_list)
        ))

    @ew.add_match(
        RenameColumns(GroupByAggregation, ...),
        lambda exp: len({src for src, _ in exp.renames} & {
            f.dst_column for f in exp.relation.aggregate_functions
        }) > 0
    )
    def merge_rename_column_group_by(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        aggregate_functions = expression.relation.aggregate_functions
        new_aggregate_functions = []
        for falm in aggregate_functions:
            if falm.dst_column in renames:
                old_dst = falm.dst_column
                falm = FunctionApplicationListMember(
                    falm.fun_exp,
                    renames[falm.dst_column]
                )
                del renames[old_dst]
            new_aggregate_functions.append(falm)
        return self.walk(
            RenameColumns(
                GroupByAggregation(
                    expression.relation.relation,
                    expression.relation.groupby,
                    tuple(new_aggregate_functions)
                ),
                tuple((src, dst) for src, dst in renames.items())
            )
        )

    @ew.add_match(
        RenameColumns(GroupByAggregation, ...),
        lambda exp: set(chain(*exp.renames)).isdisjoint(
            chain(*(
                get_expression_columns(agg.fun_exp)
                for agg in exp.relation.aggregate_functions
            ))
        )
    )
    def push_rename_past_groupby(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        new_groupby = tuple(
            renames.get(col, col) for col in expression.relation.groupby
        )

        return self.walk(
            GroupByAggregation(
                RenameColumns(
                    expression.relation.relation,
                    expression.renames
                ),
                new_groupby,
                expression.relation.aggregate_functions
            )
        )

    @ew.add_match(RenameColumns(Projection, ...))
    def push_rename_past_projection(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        new_attributes = tuple(
            renames.get(col, col) for col in expression.relation.attributes
        )

        return self.walk(
            Projection(
                RenameColumns(
                    expression.relation.relation,
                    expression.renames
                ),
                new_attributes
            )
        )

    @ew.add_match(RenameColumns(Selection, ...))
    def push_rename_past_selection(self, expression):
        selection = expression.relation
        renames = {src: dst for src, dst in expression.renames}
        rsw = ew.ReplaceExpressionWalker(renames)
        new_formula = rsw.walk(selection.formula)
        return self.walk(
            Selection(
                RenameColumns(selection.relation, expression.renames),
                new_formula
            )
        )

    @ew.add_match(RenameColumns(NameColumns, ...))
    def simplify_renames_name(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        new_names = []
        for name in expression.relation.column_names:
            name = renames.get(name, name)
            new_names.append(name)
        return self.walk(
            NameColumns(expression.relation.relation, tuple(new_names))
        )

    @ew.add_match(RenameColumns(NaturalJoin(RelationalAlgebraOperation, RelationalAlgebraOperation), ...))
    def split_rename_naturaljoin(self, expression):
        return self._rename_joinop(expression, NaturalJoin)

    @ew.add_match(RenameColumns(LeftNaturalJoin(RelationalAlgebraOperation, RelationalAlgebraOperation), ...))
    def split_rename_left_naturaljoin(self, expression):
        return self._rename_joinop(expression, LeftNaturalJoin)

    @ew.add_match(RenameColumns(FullOuterNaturalJoin(RelationalAlgebraOperation, RelationalAlgebraOperation), ...))
    def split_rename_outer_naturaljoin(self, expression):
        return self._rename_joinop(expression, FullOuterNaturalJoin)

    def _rename_joinop(self, expression, joinop):
        left_renames = []
        right_renames = []
        left_columns = expression.relation.relation_left.columns()
        right_columns = expression.relation.relation_right.columns()
        for src, dst in expression.renames:
            if src in left_columns:
                left_renames.append((src, dst))
            if src in right_columns:
                right_renames.append((src, dst))
        return self.walk(joinop(
            RenameColumns(
                expression.relation.relation_left,
                tuple(left_renames)
            ),
            RenameColumns(
                expression.relation.relation_right,
                tuple(right_renames)
            )
        ))

    @ew.add_match(RenameColumns(ReplaceNull, ...))
    def switch_rename_replace_null(self, expression):
        renames = {src: dst for src, dst in expression.renames}
        return self.walk(
            ReplaceNull(
                RenameColumns(
                    expression.relation.relation,
                    expression.renames
                ),
                renames.get(
                    expression.relation.column,
                    expression.relation.column
                ),
                expression.relation.value
            )
        )

    @ew.add_match(
        NameColumns(NumberColumns(RelationalAlgebraOperation, ...), ...),
        lambda expression: all(
            isinstance(column, Constant[ColumnStr])
            for column in expression.relation.relation.columns()
        )
    )
    def eliminate_trivial_number_columns(self, expression):
        column_renames = tuple(
            (src, dst)
            for src, dst in zip(
                expression.relation.column_names,
                expression.column_names
            )
        )

        return self.walk(
            RenameColumns(
                expression.relation.relation, column_renames
            )
        )

    @ew.add_match(
        Projection(NumberColumns(RelationalAlgebraOperation, ...), ...),
        lambda expression: all(
            isinstance(column, Constant[ColumnStr])
            for column in expression.relation.relation.columns()
        )
    )
    def eliminate_trivial_projection_number_columns(self, expression):
        attributes = tuple(
            expression.relation.relation.columns()[
                column.value
            ]
            for column in expression.relation.columns()
        )
        new_relation = Projection(
            expression.relation.relation,
            attributes
        )
        return self.walk(new_relation)

    @ew.add_match(
        Selection(NumberColumns(RelationalAlgebraOperation, ...), ...),
        lambda expression: all(
            isinstance(column, Constant[ColumnStr])
            for column in expression.relation.relation.columns()
        )
    )
    def eliminate_trivial_selection_number_columns(self, expression):
        replacements = {
            number: name for number, name in
            zip(
                expression.relation.columns(),
                expression.relation.column_names
            )
        }
        new_formula = (
            ew.ReplaceExpressionWalker(replacements)
            .walk(expression.formula)
        )
        new_relation = NumberColumns(
            Selection(expression.relation.relation, new_formula),
            expression.relation.column_names
        )

        return self.walk(new_relation)


class PushInSelections(ew.PatternWalker):
    @ew.add_match(
        Selection(NaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_right)
            ) == 0
        )
    )
    def push_selection_in_left(self, expression):
        return self.walk(
            NaturalJoin(
                Selection(
                    expression.relation.relation_left,
                    expression.formula
                ),
                expression.relation.relation_right
            )
        )

    @ew.add_match(
        Selection(NaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_left)
            ) == 0
        )
    )
    def push_selection_in_right(self, expression):
        return self.walk(
            NaturalJoin(
                expression.relation.relation_left,
                Selection(
                    expression.relation.relation_right,
                    expression.formula
                )
            )
        )

    @ew.add_match(
        Selection(
            NaturalJoin,
            FunctionApplication(EQ, (Constant[ColumnStr], Constant[ColumnStr]))
        ),
        lambda exp: (
            (
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_left)
            ) and (
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_right)
            )
        )
    )
    def push_and_rename_in_naturaljoin(self, expression):
        column_left = (
            get_expression_columns(expression.formula) &
            (
                get_expression_columns(expression.relation.relation_left) -
                get_expression_columns(expression.relation.relation_right)
            )
        ).pop()
        column_right = (
            get_expression_columns(expression.formula) &
            get_expression_columns(expression.relation.relation_right)
        ).pop()

        if column_left != column_right:
            relation_right = RenameColumn(
                expression.relation.relation_right,
                column_right, column_left
            )
            projections = tuple(
                FunctionApplicationListMember(c, c)
                for c in expression.columns()
                if c != column_right
            )
            projections += (
                FunctionApplicationListMember(column_left, column_right),
            )
            expression = ExtendedProjection(
                NaturalJoin(
                    expression.relation.relation_left,
                    relation_right
                ),
                projections
            )
        return expression

    @ew.add_match(
        Selection(LeftNaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_right)
            ) == 0
        )
    )
    def push_selection_in_leftnaturaljoin_left(self, expression):
        return self.walk(
            LeftNaturalJoin(
                Selection(
                    expression.relation.relation_left,
                    expression.formula
                ),
                expression.relation.relation_right
            )
        )

    @ew.add_match(
        Selection(LeftNaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_left)
            ) == 0
        )
    )
    def push_selection_in_leftnaturaljoin_right(self, expression):
        return self.walk(
            LeftNaturalJoin(
                expression.relation.relation_left,
                Selection(
                    expression.relation.relation_right,
                    expression.formula
                )
            )
        )

    @ew.add_match(
        Selection(FullOuterNaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_right)
            ) == 0
        )
    )
    def push_selection_in_outernaturaljoin_left(self, expression):
        return self.walk(
            FullOuterNaturalJoin(
                Selection(
                    expression.relation.relation_left,
                    expression.formula
                ),
                expression.relation.relation_right
            )
        )

    @ew.add_match(
        Selection(FullOuterNaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_left)
            ) == 0
        )
    )
    def push_selection_in_outernaturaljoin_right(self, expression):
        return self.walk(
            FullOuterNaturalJoin(
                expression.relation.relation_left,
                Selection(
                    expression.relation.relation_right,
                    expression.formula
                )
            )
        )

    @ew.add_match(
        Selection(Projection, ...),
        lambda exp: (
            all(
                isinstance(column, Constant[ColumnStr])
                for column in get_expression_columns(exp.formula)
            ) and
            get_expression_columns(exp.formula) <=
            set(exp.relation.attributes)
        )
    )
    def push_selection_in_projection(self, expression):
        return self.walk(Projection(
            Selection(expression.relation.relation, expression.formula),
            expression.relation.attributes
        ))

    @ew.add_match(
        Selection(ExtendedProjection, ...),
        lambda exp: (
            get_expression_columns(exp.formula) <=
            {
                projection.dst_column
                for projection in exp.relation.projection_list
                if isinstance(projection.fun_exp, Constant[Column])
            }
        )
    )
    def push_selection_in_extended_projection(self, expression):
        rew = ew.ReplaceExpressionWalker({
            projection.dst_column: projection.fun_exp
            for projection in expression.relation.projection_list
            if isinstance(projection.fun_exp, Constant[Column])
        })
        new_formula = rew.walk(expression.formula)
        return self.walk(ExtendedProjection(
            Selection(expression.relation.relation, new_formula),
            expression.relation.projection_list
        ))

    @ew.add_match(
        Selection(GroupByAggregation, ...),
        lambda exp: (
            get_expression_columns(exp.formula) <=
            set(exp.relation.groupby)
        )
    )
    def push_selection_in_groupby(self, expression):
        return self.walk(
            GroupByAggregation(
                Selection(
                    expression.relation.relation,
                    expression.formula
                ),
                expression.relation.groupby,
                expression.relation.aggregate_functions
            )
        )

    @ew.add_match(
        Selection(ReplaceNull, ...),
        lambda exp: (
            exp.relation.column not in get_expression_columns(exp.formula)
        )
    )
    def push_selection_in_replace_null(self, expression):
        return ReplaceNull(
            Selection(
                expression.relation.relation,
                expression.formula
            ),
            expression.relation.column,
            expression.relation.value
        )


class PushUnnamedSelectionsUp(ew.PatternWalker):

    @staticmethod
    def _extract_column_int_constants(expression):
        return {
            e for _, e in ew.expression_iterator(expression)
            if isinstance(e, Constant[ColumnInt])
        }

    @ew.add_match(
        Projection(
            Selection(..., eq_(Constant[ColumnInt], Constant[ColumnInt])),
            ...
        ),
        lambda expression: (
            max(expression.relation.formula.args, key=lambda x: x.value)
            in set(expression.attributes)
        ) and (
            min(expression.relation.formula.args, key=lambda x: x.value)
            not in set(expression.attributes)
        )
    )
    def standardize_projected_column(self, expression):
        rew = ew.ReplaceExpressionWalker(
            {
                max(expression.relation.formula.args, key=lambda x: x.value):
                min(expression.relation.formula.args, key=lambda x: x.value)
            }
        )

        return Projection(
            expression.relation,
            rew.walk(expression.attributes)
        )

    @ew.add_match(
        Selection(Selection, eq_(Constant[ColumnInt], Constant[ColumnInt])),
        lambda expression: not (
            isinstance(expression.relation.formula, FunctionApplication) and
            expression.relation.formula.functor == eq_ and
            isinstance(
                expression.relation.formula.args[0],
                Constant[ColumnInt]
            ) and
            isinstance(
                expression.relation.formula.args[1],
                Constant[ColumnInt]
            )
        )
    )
    def invert_selection_order(self, expression):
        rew = ew.ReplaceExpressionWalker(
            {
                max(expression.formula.args, key=lambda x: x.value):
                min(expression.formula.args, key=lambda x: x.value)
            }
        )
        return Selection(
            Selection(
                expression.relation.relation,
                expression.formula
            ),
            rew.walk(expression.relation.formula)
        )

    @ew.add_match(
        Projection(Selection, ...),
        lambda expression: (
            (
                PushUnnamedSelectionsUp
                ._extract_column_int_constants(expression.relation.formula)
            ) and (
                PushUnnamedSelectionsUp
                ._extract_column_int_constants(expression.relation.formula)
                <= set(expression.attributes)
            )
        )
    )
    def push_selection_above_projection(self, expression):
        return Selection(
            Projection(
                expression.relation.relation,
                expression.attributes
            ),
            expression.relation.formula
        )

    @ew.add_match(
        NameColumns(Selection, ...),
        lambda expression: (
            PushUnnamedSelectionsUp
            ._extract_column_int_constants(expression.relation.formula)
        )
    )
    def push_selection_above_name_columns(self, expression):
        rew = ew.ReplaceExpressionWalker({
            int2columnint_constant(i): name
            for i, name in enumerate(expression.column_names)
        })
        new_formula = rew.walk(expression.relation.formula)
        return Selection(
            NameColumns(
                expression.relation.relation,
                expression.column_names
            ),
            new_formula
        )


class CommuteJoinsAvoidCrossProducts(ew.PatternWalker):
    @ew.add_match(
        NaturalJoin(..., NaturalJoin),
        lambda exp: (
            exp.relation_left.columns() & exp.relation_right.columns() and
            not exp.relation_right.relation_left.columns() & exp.relation_right.relation_right.columns()
        )
    )
    def commute_right(self, expression):
        left, right = expression.unapply()
        right_left, right_right = right.unapply()

        if left.columns() & right_right.columns():
            res = expression.apply(
                right_left,
                expression.apply(left, right_right)
            )
        elif left.columns() & right_left.columns():
            res = expression.apply(
                right_right,
                expression.apply(left, right_left)
            )
        return res

    @ew.add_match(
        NaturalJoin(NaturalJoin, ...),
        lambda exp: (
            exp.relation_left.columns() & exp.relation_right.columns() and
            not exp.relation_left.relation_left.columns() & exp.relation_left.relation_right.columns()
        )
    )
    def commute_left(self, expression):
        left, right = expression.unapply()
        left_left, left_right = left.unapply()

        if right.columns() & left_right.columns():
            res = expression.apply(
                left_left,
                expression.apply(right, left_right)
            )
        elif right.columns() & left_left.columns():
            res = expression.apply(
                left_right,
                expression.apply(right, left_left)
            )
        return res


class RelationalAlgebraOptimiser(
    RewriteSelections,
    ProductSimplification,
    EliminateTrivialProjections,
    PushInSelections,
    RenameOptimizations,
    PushUnnamedSelectionsUp,
    CommuteJoinsAvoidCrossProducts,
    ew.ExpressionWalker,
):
    """
    Mixing that optimises through relational algebra expressions by
    rewriting.
    equi-selection/product compositions into equijoins.
    """

    pass



import base64
import hashlib
from io import BytesIO
import json
from concurrent.futures import Future
from collections import namedtuple

import matplotlib
from neurolang.frontend.query_resolution_expressions import Symbol
from typing import Any, Dict, List, Tuple, Type, Union
from nibabel.nifti1 import Nifti1Image
from nibabel.spatialimages import SpatialImage

import numpy as np
import pandas as pd
import nibabel as nib
from neurolang.regions import EmptyRegion, ExplicitVBR, ExplicitVBROverlay
from neurolang.type_system import get_args
from neurolang.utils.relational_algebra_set import (
    RelationalAlgebraFrozenSet,
)
from neurolang.exceptions import ParserError


def base64_encode_nifti(image):
    """Returns base64 encoded string of the specified image.

    Parameters
    ----------
    image : nibabel.Nifti2Image
        image to be encoded.

    Returns
    -------
    str
        base64 encoded string of the image.
    """
    encoded_image = base64.encodebytes(image.to_bytes())
    enc = encoded_image.decode("utf-8")
    return enc


def base64_encode_spatial(image: SpatialImage):
    """Returns base64 encoded string of a spatial image

    Parameters
    ----------
    image : nibabel.spatialimages.SpatialImage
        spatial image to be encoded.

    Returns
    -------
    str
        base64 encoded string of the vbr.
    """
    nifti_image = Nifti1Image(
        np.asanyarray(image.dataobj, dtype=np.float32), affine=image.affine
    )
    return base64_encode_nifti(nifti_image)


def calculate_image_center(image: SpatialImage):
    """Calculates center coordinates for the specified image."""
    coords = np.transpose(image.get_fdata().nonzero()).mean(0).astype(int)
    coords = nib.affines.apply_affine(image.affine, coords)
    return [int(c) for c in coords]


def serializeVBR(image_row: pd.Series):
    """
    Serialize a Volumetric Brain Region object.

    Parameters
    ----------
    vbr : Union[ExplicitVBR, ExplicitVBROverlay]
        the volumetric brain region to serialize

    Returns
    -------
    Dict
        a dict containing the base64 encoded image, as well as min and max
        values, and a hash of the image.
    """
    index = image_row["index"]
    vbr = image_row[1]
    if isinstance(vbr, EmptyRegion):
        return "Empty Region"

    image = vbr.spatial_image()
    flattened = image.get_fdata().flatten()
    min = flattened[flattened != 0].min()
    max = flattened.max()
    q95 = np.quantile(flattened[flattened != 0], 0.95)
    hash = hashlib.sha224(image.dataobj.tobytes()).hexdigest()
    return {
        "min": min,
        "max": max,
        "q95": q95,
        "image": base64_encode_spatial(image),
        "hash": hash,
        "center": calculate_image_center(image),
        "idx": index,
    }


def serialize_mplt_thumbnails(figure: matplotlib.figure.Figure):
    """
    Serialize a figure by creating a base64 encoded thumbnail.
    This is used only to display the figure as a thumbnail in the list of
    results. The full size figure will be displayed when the user clicks on
    the thumbnail. See `app.MpltFigureHandler`.

    Parameters
    ----------
    figure : matplotlib.figure.Figure
        the figure to serialize

    Returns
    -------
    str
        base64 encoded thumbnail of the figure
    """
    data = BytesIO()
    # set dpi to low value to create a small image
    figure.savefig(data, format="png", dpi=10)
    data.seek(0)
    encoded_figure = base64.encodebytes(data.read())
    enc = encoded_figure.decode("utf-8")
    return enc


class CustomQueryResultsEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, QueryResults):
            return obj.__dict__
        return super().default(obj)


class QueryResults:
    """
    A representation of query results. This class is returned by
    the tornado application as a JSON serialized string.

    It contains
        * metadata information about the query:
            - cancelled, running, done
        * metadata about the requested information:
            - the specific symbol to return, as well as
            - start, length, sort, asc (parameters for which rows to return)
        * if the query resulted in an error, the error details
        * if the query resulted in data, the details for each requested symbol:
            - its columns, size, & row_type
            - the rows corresponding to the start, length, sort & asc params
    """

    def __init__(
        self,
        uuid: str,
        future: Future,
        symbol: str = None,
        start: int = 0,
        length: int = 50,
        sort: int = -1,
        asc: bool = True,
        get_values: bool = False,
    ):
        self.start = start
        self.length = length
        self.sort = sort
        self.asc = asc
        # First, set the uuid and status of the future
        self.uuid = uuid
        self.cancelled = future.cancelled()
        self.running = future.running()
        self.done = future.done()

        # Then, check if it's done
        if future.done():
            error = future.exception()
            if error is not None:
                self.set_error_details(error)
            else:
                results = future.result()
                if results is not None:
                    self.set_results_details(results, symbol, get_values)

    def set_error_details(self, error):
        self.errorName = str(type(error))
        if isinstance(error, ParserError):
            self.message = "An error occured while parsing your query."
            self.errorDoc = str(error)
            if hasattr(error, 'tokenizer'):
                line_info = error.tokenizer.line_info(error.pos)
            elif hasattr(error, 'buf'):
                line_info = error.buf.line_info(error.pos)
            elif hasattr(error, 'line') and hasattr(error, 'column'):
                LineInfo = namedtuple("LineInfo", ["line", "col"])
                line_info = LineInfo(line=error.line, col=error.column)
            self.line_info = {
                "line": line_info.line,
                "col": line_info.col,
                "text": error.message,
            }
        else:
            self.message = str(error)
            if error.__doc__ is not None:
                self.errorDoc = error.__doc__

    def set_results_details(self, results, symbol, get_values):
        self.results = {}
        if symbol is None:
            for key, ras in results.items():
                self.results[key] = self.get_result_item(ras, get_values)
        else:
            self.results[symbol] = self.get_result_item(results[symbol], True)

    def get_result_item(
        self,
        symbol: Union[RelationalAlgebraFrozenSet, Symbol],
        get_item_values: bool = False,
    ) -> Dict:
        """
        Serialize a symbol into a dict of result values and metadata.

        Parameters
        ----------
        symbol : Union[RelationalAlgebraFrozenSet, Symbol]
            the symbol to parse
        get_item_values : bool
            get the values for the item

        Returns
        -------
        Dict
            the parsed result values
        """
        if isinstance(symbol, RelationalAlgebraFrozenSet):
            df = symbol.as_pandas_dataframe()
            result = self.get_result_item_columns(symbol, df)
            if get_item_values:
                values = self.get_result_item_values(symbol.row_type, df)
                result["values"] = values
        elif isinstance(symbol, Symbol):
            result = self.get_function_metadata(symbol)
        elif isinstance(symbol, dict):
            result = symbol
        return result

    def get_function_metadata(self, symbol: Symbol):
        result = {
            "type": str(symbol.type),
            "doc": symbol.value.__doc__,
            "function": True,
        }
        return result

    def get_result_item_columns(
        self, symbol: RelationalAlgebraFrozenSet, df
    ) -> Dict:
        result = {}
        result["row_type"] = [str(t) for t in get_args(symbol.row_type)]
        result["columns"] = [str(c) for c in symbol.columns]
        result["size"] = df.shape[0]
        result["probabilistic"] = (
            hasattr(symbol, "_is_probabilistic") and symbol._is_probabilistic
        )
        result["last_parsed_symbol"] = (
            hasattr(symbol, "_last_parsed_symbol")
            and symbol._last_parsed_symbol
        )
        return result

    def get_result_item_values(
        self, row_type: Union[Any, Type[Tuple], None], df: pd.DataFrame
    ) -> List[List]:
        """
        Return the rows of the dataframe corresponding to the requested slice
        (as specified by start, length & sort values) in a json compatible form.

        Parameters
        ----------
        row_type : Union[Any, Type[Tuple],  None]
            the row_type.
        df : pd.DataFrame
            the dataframe.

        Returns
        -------
        List[List]
            the values.
        """
        if self.sort > -1:
            df = df.sort_values(by=[df.columns[self.sort]], ascending=self.asc)
        rows = df.iloc[self.start : self.start + self.length].copy()
        for col, col_type in zip(rows.columns, row_type.__args__):
            if col_type == ExplicitVBR or col_type == ExplicitVBROverlay:
                rows[col] = (
                    rows[col].reset_index().apply(serializeVBR, axis=1).values
                )
            elif col_type == matplotlib.figure.Figure:
                rows[col] = rows[col].apply(serialize_mplt_thumbnails)
            elif rows[col].dtype == np.object_:
                rows[col] = rows[col].astype(str)

        return rows.values.tolist()



"""
This module exposes utility functions for tests on logic expressions.

It should not be used for any other purpose than testing.

"""
from ...expression_pattern_matching import add_match
from ...expression_walker import ExpressionWalker
from ...expressions import Definition, Expression
from ...logic import NaryLogicOperator

__all__ = [
    "logic_exp_commutative_equal",
]


class LogicCommutativeComparison(Definition):
    """
    Comparison between two expressions that uses the commutativity property of
    some logic operators such as conjunctions and disjunctions.

    Parameters
    ----------
    first : Expression
        First expression.

    second : Expression
        Second expression.

    """

    def __init__(self, first, second):
        self.first = first
        self.second = second

    def __repr__(self):
        return "Compare\n\t{}\nwith\n\t{}".format(
            repr(self.first), repr(self.second)
        )


class LogicCommutativeComparator(ExpressionWalker):
    """
    Compare logic expressions using the commutativity property of some logic
    operators such as conjunctions and disjunctions.

    """

    @add_match(
        LogicCommutativeComparison(NaryLogicOperator, NaryLogicOperator)
    )
    def nary_logic_operators(self, comp):
        """
        Compare two n-ary logic operators by comparing their two sets of
        formulas.

        """
        if not isinstance(comp.first, type(comp.second)) or not isinstance(
            comp.second, type(comp.first)
        ):
            return False
        return self._compare_set_of_formulas(comp.first, comp.second)

    @add_match(LogicCommutativeComparison(Expression, Expression))
    def expressions(self, comp):
        args1 = comp.first.unapply()
        args2 = comp.second.unapply()
        if len(args1) != len(args2):
            return False
        for arg1, arg2 in zip(args1, args2):
            if not self._args_equal(arg1, arg2):
                return False
        return True

    def _args_equal(self, arg1, arg2):
        if isinstance(arg1, Expression) and isinstance(arg2, Expression):
            if not self.walk(LogicCommutativeComparison(arg1, arg2)):
                return False
        elif arg1 != arg2:
            return False
        return True

    def _compare_set_of_formulas(self, first, second):
        return all(
            any(
                self.walk(LogicCommutativeComparison(f1, f2))
                for f2 in second.formulas
            )
            for f1 in first.formulas
        )


def logic_exp_commutative_equal(exp1, exp2):
    """
    Compare two expressions using the commutativity property of logic
    operators.

    The two expressions do not need to be purely equal if the order of the
    formulas of a commutative logic operator is not the same in the two
    expressions.

    Apart from commutative logic operators, the comparison between the two
    expressions remains the same as the equality comparison.

    Parameters
    ----------
    exp1 : Expression
        First expression.

    exp2 : Expression
        Second expression.

    """
    if not isinstance(exp1, Expression) or not isinstance(exp2, Expression):
        raise ValueError("Can only compare expressions")
    return LogicCommutativeComparator().walk(
        LogicCommutativeComparison(exp1, exp2)
    )



import logging
import operator
import typing
from typing import AbstractSet, Tuple

from ..datalog.aggregation import is_builtin_aggregation_functor
from ..datalog.expression_processing import (
    EQ,
    conjunct_formulas,
    extract_logic_free_variables
)
from ..datalog.instance import MapInstance, WrappedRelationalAlgebraFrozenSet
from ..expression_pattern_matching import add_match
from ..expression_walker import (
    ChainedWalker,
    ExpressionWalker,
    PatternWalker,
)
from ..expressions import Constant, Expression, FunctionApplication, Symbol
from ..logic import TRUE, Conjunction, Implication, Union
from ..relational_algebra import (
    EliminateTrivialProjections,
    ExtendedProjection,
    FunctionApplicationListMember,
    Projection,
    PushInSelections,
    RelationalAlgebraOperation,
    RelationalAlgebraSolver,
    RenameOptimizations,
    Selection,
    SimplifyExtendedProjectionsWithConstants,
    str2columnstr_constant
)
from ..relational_algebra.optimisers import PushUnnamedSelectionsUp
from ..relational_algebra_provenance import (
    NaturalJoinInverse,
    ProvenanceAlgebraSet
)
from .cplogic.program import CPLogicProgram
from .exceptions import RepeatedTuplesInProbabilisticRelationError
from .expression_processing import (
    construct_within_language_succ_result,
    is_query_based_probpredicate,
    is_within_language_prob_query,
    within_language_succ_query_to_intensional_rule
)
from .expressions import Condition, ProbabilisticPredicate
from .probabilistic_semiring_solver import (
    ProbSemiringToRelationalAlgebraSolver
)

AND = Constant(operator.and_)
NE = Constant(operator.ne)

ZERO = Constant[float](0.)
GT = Constant(operator.gt)


def _qbased_probfact_needs_translation(formula: Implication) -> bool:
    if isinstance(formula.antecedent, FunctionApplication):
        antecedent_pred = formula.antecedent
    elif (
        isinstance(formula.antecedent, Conjunction)
        and len(formula.antecedent.formulas) == 1
    ):
        antecedent_pred = formula.antecedent.formulas[0]
    else:
        return True
    return not (
        isinstance(antecedent_pred.functor, Symbol)
        and antecedent_pred.functor.is_fresh
    )


LOG = logging.getLogger(__name__)


class QueryBasedProbFactToDetRule(PatternWalker):
    """
    Translate a query-based probabilistic fact to two rules. A deterministic
    rule that takes care of inferring the set of probabilistic tuples and their
    probabilities and a probabilistic rule that transforms the result from the
    deterministic resolution into a proper probabilistic table by re-attaching
    the deterministically-obtained probabilities.

    For example, the rule `P(x) : f(x) :- Q(x)` is translated into the rules

        _f1_(_f2_, x) :- Q(x), _f2_ = f(x)
        P(x) : _f2_ :- _f1_(_f2_, x)

    where `_f1_` and `_f2_` are fresh symbols. Note: we make sure not to
    expose the resulting `_f1_` relation to the user, as it is not part of
    her program.

    """

    @add_match(
        Union,
        lambda union: any(
            is_query_based_probpredicate(formula)
            and _qbased_probfact_needs_translation(formula)
            for formula in union.formulas
        ),
    )
    def union_with_query_based_pfact(self, union):
        new_formulas = list()
        for formula in union.formulas:
            if is_query_based_probpredicate(
                formula
            ) and _qbased_probfact_needs_translation(formula):
                (
                    det_rule,
                    prob_rule,
                ) = self._query_based_probabilistic_fact_to_det_and_prob_rules(
                    formula
                )
                new_formulas.append(det_rule)
                new_formulas.append(prob_rule)
            else:
                new_formulas.append(formula)
        return self.walk(Union(tuple(new_formulas)))

    @add_match(
        Implication,
        lambda implication: is_query_based_probpredicate(implication)
        and _qbased_probfact_needs_translation(implication),
    )
    def query_based_probafact(self, impl):
        (
            det_rule,
            prob_rule,
        ) = self._query_based_probabilistic_fact_to_det_and_prob_rules(impl)
        return self.walk(Union((det_rule, prob_rule)))

    @staticmethod
    def _query_based_probabilistic_fact_to_det_and_prob_rules(impl):
        prob_symb = Symbol.fresh()
        det_pred_symb = Symbol.fresh()
        agg_functor = QueryBasedProbFactToDetRule._get_agg_functor(impl)
        if agg_functor is None:
            eq_formula = EQ(prob_symb, impl.consequent.probability)
            det_consequent = det_pred_symb(
                prob_symb, *impl.consequent.body.args
            )
            prob_antecedent = det_consequent
        else:
            assert len(impl.consequent.probability.args) == 1
            eq_formula = EQ(prob_symb, impl.consequent.probability.args[0])
            det_consequent = det_pred_symb(
                agg_functor(prob_symb), *impl.consequent.body.args
            )
            prob_antecedent = det_pred_symb(
                prob_symb, *impl.consequent.body.args
            )
        det_antecedent = conjunct_formulas(impl.antecedent, eq_formula)
        det_rule = Implication(det_consequent, det_antecedent)
        prob_consequent = impl.consequent.apply(
            prob_symb, impl.consequent.body
        )
        prob_rule = Implication(prob_consequent, prob_antecedent)
        return det_rule, prob_rule

    @staticmethod
    def _get_agg_functor(
        impl: Implication,
    ) -> typing.Union[None, typing.Callable]:
        if not isinstance(
            impl.consequent.probability, FunctionApplication
        ) or not is_builtin_aggregation_functor(
            impl.consequent.probability.functor
        ):
            return
        return impl.consequent.probability.functor


def _solve_within_language_prob_query(
    cpl: CPLogicProgram,
    rule: Implication,
    succ_prob_solver: typing.Callable,
    marg_prob_solver: typing.Callable,
) -> Constant[AbstractSet]:
    query = within_language_succ_query_to_intensional_rule(rule)
    if isinstance(rule.antecedent, Condition):
        provset = marg_prob_solver(query, cpl)
    else:
        provset = succ_prob_solver(query, cpl)
    relation = construct_within_language_succ_result(provset, rule)
    return relation


def _solve_for_probabilistic_rule(
    cpl: CPLogicProgram,
    rule: Implication,
    succ_prob_solver: typing.Callable,
):
    provset = succ_prob_solver(rule, cpl)
    return provset.relation


def compute_probabilistic_solution(
    det_edb,
    pfact_db,
    pchoice_edb,
    prob_idb,
    succ_prob_solver,
    marg_prob_solver,
    check_qbased_pfact_tuple_unicity=False,
):
    solution = MapInstance()
    cpl, prob_idb = _build_probabilistic_program(
        det_edb,
        pfact_db,
        pchoice_edb,
        prob_idb,
        check_qbased_pfact_tuple_unicity,
    )
    for rule in prob_idb.formulas:
        if is_within_language_prob_query(rule):
            relation = _solve_within_language_prob_query(
                cpl, rule, succ_prob_solver, marg_prob_solver
            )
            solution[rule.consequent.functor] = Constant[AbstractSet](
                relation.value.to_unnamed()
            )
    return solution


def lift_solve_marg_query(rule, cpl, succ_solver):
    """
    Solve a MARG query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        Consequent must be of type `Condition`.
        MARG query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """
    res_args = tuple(s for s in rule.consequent.args if isinstance(s, Symbol))

    joint_antecedent = Conjunction(
        (
            rule.antecedent.conditioned,
            rule.antecedent.conditioning
        )
    )
    joint_logic_variables = set(res_args)
    joint_rule = Implication(
        Symbol.fresh()(*joint_logic_variables), joint_antecedent
    )
    joint_provset = succ_solver(
        joint_rule, cpl, run_relational_algebra_solver=False
    )

    denominator_antecedent = rule.antecedent.conditioning
    denominator_logic_variables = (
        extract_logic_free_variables(denominator_antecedent) & res_args
    )
    denominator_rule = Implication(
        Symbol.fresh()(*denominator_logic_variables), denominator_antecedent
    )
    denominator_provset = succ_solver(
        denominator_rule, cpl, run_relational_algebra_solver=False
    )
    query_solver = generate_provenance_query_solver({}, True)
    provset = query_solver.walk(
        Projection(
            NaturalJoinInverse(joint_provset, denominator_provset),
            tuple(str2columnstr_constant(s.name) for s in res_args),
        )
    )
    return provset


def _discard_query_based_probpredicates(prob_idb):
    return Union(
        tuple(
            formula
            for formula in prob_idb.formulas
            if not (
                isinstance(formula.consequent, ProbabilisticPredicate)
                and formula.antecedent != TRUE
            )
        )
    )


def _add_to_probabilistic_program(
    add_fun,
    pred_symb,
    expr,
    det_edb,
    check_qbased_pfact_tuple_unicity=False,
):
    # handle set-based probabilistic tables
    if isinstance(expr, Constant[typing.AbstractSet]):
        ra_set = expr.value
    # handle query-based probabilistic facts
    elif isinstance(expr, Union):
        impl = expr.formulas[0]
        # we know the rule is of the form
        # P(x_1, ..., x_n) : y :- Q(y, x_1, ..., x_n)
        # where Q is an extensional relation symbol
        # so the values can be retrieved from the EDB
        if impl.antecedent.functor in det_edb:
            ra_set = det_edb[impl.antecedent.functor].value
            if check_qbased_pfact_tuple_unicity:
                _check_tuple_prob_unicity(ra_set)
        else:
            ra_set = WrappedRelationalAlgebraFrozenSet.dum()
    add_fun(pred_symb, ra_set.unwrap())


def _check_tuple_prob_unicity(ra_set: Constant[AbstractSet]) -> None:
    length = len(ra_set)
    proj_cols = list(ra_set.columns)[1:]
    length_without_probs = len(ra_set.projection(*proj_cols))
    if length_without_probs != length:
        n_repeated_tuples = length - length_without_probs
        raise RepeatedTuplesInProbabilisticRelationError(
            n_repeated_tuples,
            length,
            "Some tuples have multiple probability labels. "
            f"Found {n_repeated_tuples} tuple repetitions, out of "
            f"{length} total tuples. If your query-based probabilistic fact "
            "leads to multiple probabilities for the same tuple, you might "
            "want to consider aggregating these probabilities by taking their "
            "maximum or average.",
        )


def _build_probabilistic_program(
    det_edb,
    pfact_db,
    pchoice_edb,
    prob_idb,
    check_qbased_pfact_tuple_unicity=False,
):
    cpl = CPLogicProgram()
    db_to_add_fun = [
        (det_edb, cpl.add_extensional_predicate_from_tuples),
        (pfact_db, cpl.add_probabilistic_facts_from_tuples),
        (pchoice_edb, cpl.add_probabilistic_choice_from_tuples),
    ]
    for database, add_fun in db_to_add_fun:
        for pred_symb, expr in database.items():
            _add_to_probabilistic_program(
                add_fun,
                pred_symb,
                expr,
                det_edb,
                check_qbased_pfact_tuple_unicity,
            )
    # remove query-based probabilistic facts that have already been processed
    # and transformed into probabilistic tables based on the deterministic
    # solution of their probability and antecedent
    prob_idb = _discard_query_based_probpredicates(prob_idb)
    cpl.walk(prob_idb)
    return cpl, prob_idb


class FloatArithmeticSimplifier(PatternWalker):
    @add_match(
        FunctionApplication(
            Constant[typing.Any](operator.mul),
            (Constant[float](1.0), Expression[typing.Any]))
        )
    def simplify_mul_left(self, expression):
        return self.walk(expression.args[1])

    @add_match(
        FunctionApplication(
            Constant[typing.Any](operator.mul),
            (Expression[typing.Any], Constant[float](1.0))
        )
    )
    def simplify_mul_right(self, expression):
        return self.walk(expression.args[0])


class RAQueryOptimiser(
    EliminateTrivialProjections,
    PushInSelections,
    RenameOptimizations,
    SimplifyExtendedProjectionsWithConstants,
    FloatArithmeticSimplifier,
    PushUnnamedSelectionsUp,
    ExpressionWalker,
):
    pass


class AddNeededProjections(PatternWalker):
    def __init__(self, needed_projections):
        self.needed_projections = needed_projections

    @add_match(ProvenanceAlgebraSet)
    def add_projection(self, expression):
        if self.needed_projections:
            projections = (
                self.needed_projections +
                (FunctionApplicationListMember(
                    expression.provenance_column,
                    expression.provenance_column
                ),)
            )
            expression = ProvenanceAlgebraSet(
                ExtendedProjection(
                    expression.relation,
                    projections
                ),
                expression.provenance_column
            )
        return expression


class FilterZeroProbability(ExpressionWalker):
    @add_match(ProvenanceAlgebraSet)
    def add_zero_filter(self, expression):
        return ProvenanceAlgebraSet(
            Selection(
                expression.relation,
                GT(expression.provenance_column, ZERO)
            ),
            expression.provenance_column
        )


def generate_provenance_query_solver(
    symbol_table, run_relational_algebra_solver,
    needed_projections=None,
    solver_class=ProbSemiringToRelationalAlgebraSolver
):
    """
    Generate a walker that solves a RAP query.

    Parameters
    ----------
    symbol_table : Mapping
        Mapping from symbols to probabilistic or deterministic sets to
        solve the query.
    run_relational_algebra_solver : bool
        if `true` the walker will return a ProvenanceAlgebraSet containing
        a NamedAlgebraSet as `relation` attribute. If `false` the walker will
        produce a relational algebra expression as `relation` attribute.
    solver_class: PatternWalker
        class to translate a provenance RA sets program into a RA program.
        Default is `ProbSemiringToRelationalAlgebraSolver`.
    """

    if needed_projections is None:
        needed_projections = tuple()

    class LogExpression(PatternWalker):
        def __init__(self, logger, message, level):
            self.logger = logger
            self.message = message
            self.level = level

        @add_match(...)
        def log_exp(self, expression):
            LOG.log(self.level, self.message, expression)
            return expression

    steps = [
        RAQueryOptimiser(),
        solver_class(symbol_table=symbol_table),
        LogExpression(LOG, "About to optimise RA query %s", logging.INFO),
        AddNeededProjections(needed_projections),
        FilterZeroProbability(),
        RAQueryOptimiser(),
        LogExpression(LOG, "Optimised RA query %s", logging.INFO)
    ]

    if run_relational_algebra_solver:
        steps.append(RelationalAlgebraSolver())

    query_compiler = ChainedWalker(*steps)
    return query_compiler


def compute_projections_needed_to_reintroduce_head_terms(
    ra_query: RelationalAlgebraOperation,
    flat_query: Implication,
    unified_query: Implication,
) -> Tuple[FunctionApplicationListMember, ...]:
    """
    Produce the extended projection list for the terms that have been
    removed through unification of the query.

    There are two such cases:

    1. A head variable was repeated, such as in `ans(y, y)` and the extensional
    plan computes all possible values for `y` but will output only one column.
    This function makes sure that a second column for `y` is outputed, and that
    the resulting solution set will be a binary relation, as required.

    2. The query's head contains a constant, such as in `ans(2, y)`, in which
    case a constant column is added to the solution set.

    """
    proj_list = list()
    needed = False
    for old, new in zip(
        flat_query.consequent.args, unified_query.consequent.args
    ):
        dst_column = str2columnstr_constant(old.name)
        fun_exp = dst_column
        if new != old:
            if isinstance(new, Symbol):
                fun_exp = str2columnstr_constant(new.name)
            elif isinstance(new, Constant):
                fun_exp = new
            else:
                raise ValueError(
                    f"Unexpected argument {new}. "
                    "Expected symbol or constant"
                )
            needed = True
        member = FunctionApplicationListMember(fun_exp, dst_column)
        proj_list.append(member)
    if not needed:
        proj_list = []
    return tuple(proj_list)



import configparser
import distutils
import os
import sys
import shutil
import subprocess
from setuptools import find_packages, setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.build_py import build_py
from pathlib import Path

# Workaround for editable install with pip
# and versioneer
sys.path.append(os.path.dirname(__file__))

import versioneer


def update_config_file():
    """
    Read the config file in `neurolang/config/config.ini`
    and update the value for ["RAS"]["Backend"] to dask.
    This config file will then be copied during install as a
    setuptools data_files.
    """
    config_file = Path("neurolang/config/config.ini")
    config = configparser.ConfigParser(
        allow_no_value=True, comment_prefixes="//"
    )
    config.optionxform = str
    config.read(config_file)
    config["RAS"]["backend"] = "dask"
    with open(config_file, "w") as configfile:
        config.write(configfile)


class DevelopCommand(develop):
    """
    Custom develop command.

    - Option to edit the config file and set
    dask-sql as the backend when `--dask` flag is set on install.

    Usage
    -----
    Run
    $ python setup.py develop --dask
    or
    $ pip install -e . --install-option='--dask'

    - Option to build the frontend app using the npm_build command

    Usage
    -----
    Run
    $ python setup.py develop --npm-build
    or
    $ pip install -e . --install-option='--npm-build'

    This will only work when installing a source distribution zip or tarball,
    or installing in editable mode from a source tree. **It will not work when
    installing from a binary wheel.**
    """

    description = "develop option to set dask-sql as backend"

    user_options = develop.user_options + [
        ("dask", None, None),
        ("npm-build", None, "build the frontend html/js app using npm-build"),
    ]

    def initialize_options(self):
        develop.initialize_options(self)
        self.dask = None
        self.npm_build = False

    def finalize_options(self):
        develop.finalize_options(self)

    def run(self):
        if self.dask:
            update_config_file()
        if self.npm_build:
            self.run_command("npm_build")
        develop.run(self)


class InstallCommand(install):
    """
    Custom install command.

    - Option to edit the config file and set
    dask-sql as the backend when `--dask` flag is set on install.

    Usage
    -----
    Run
    $ python setup.py install --dask
    or
    $ pip install . --install-option='--dask'

    - Option to build the frontend app using the npm_build command

    Usage
    -----
    Run
    $ python setup.py install --npm-build
    or
    $ pip install . --install-option='--npm-build'

    This will only work when installing a source distribution zip or tarball,
    or installing in editable mode from a source tree. **It will not work when
    installing from a binary wheel.**
    """

    description = "install option to set dask-sql as backend"

    user_options = install.user_options + [
        ("dask", None, None),
        ("npm-build", None, "build the frontend html/js app using npm-build"),
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.dask = None
        self.npm_build = False

    def finalize_options(self):
        install.finalize_options(self)

    def run(self):
        if self.dask:
            update_config_file()
        if self.npm_build:
            self.run_command("npm_build")
        install.run(self)


class NPMBuildCommand(distutils.cmd.Command):
    """Run the npm build command"""

    description = "run the npm build command for the web server"
    user_options = [
        ("force", "f", "force npm build even if dist directory already exists")
    ]

    def initialize_options(self):
        """Set default values for options."""
        self.force = False

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        """Run the npm build command"""
        npm_command = shutil.which("npm")
        if not npm_command:
            raise OSError(
                "Could not find the npm binary required to build the frontend "
                "app. NPM is Node.js' package manager and can be installed "
                "with node from https://nodejs.org/en/download/."
            )

        web_dir = (
            Path(__file__).resolve().parent
            / "neurolang"
            / "utils"
            / "server"
            / "neurolang-web"
        )
        if not web_dir.exists():
            raise OSError(f"{web_dir} directory does not exist.")

        dist_dir = web_dir / "dist"
        if dist_dir.exists() and not self.force:
            self.announce(
                f"Build directory already exists for frontend app. Not rebuilding.",
                level=distutils.log.INFO,
            )
            return

        command = [npm_command, "install"]
        self.announce(
            f"Running command: [{web_dir}]$ {' '.join(command)}",
            level=distutils.log.INFO,
        )
        subprocess.check_call(
            command, cwd="neurolang/utils/server/neurolang-web"
        )

        command = [npm_command, "run", "build", "--", "--mode", "dev"]
        self.announce(
            f"Running command: [{web_dir}]$ {' '.join(command)}",
            level=distutils.log.INFO,
        )
        subprocess.check_call(
            command, cwd="neurolang/utils/server/neurolang-web"
        )


class BuildPyCommand(build_py):
    """Customize the build command and add the npm build"""

    def run(self):
        """Run the npm build before the normal build"""
        try:
            self.run_command("npm_build")
        except (OSError, subprocess.CalledProcessError) as e:
            self.announce(
                f"Skipping build of frontend app because: {e}.",
                level=distutils.log.WARN,
            )
        build_py.run(self)


if __name__ == "__main__":
    version = (versioneer.get_version(),)
    cmdclass = (versioneer.get_cmdclass(),)
    setup(
        use_scm_version=True,
        packages=find_packages(),
        cmdclass={
            "install": InstallCommand,
            "develop": DevelopCommand,
            "npm_build": NPMBuildCommand,
            "build_py": BuildPyCommand,
        },
    )
